use core::fmt::Debug;

pub struct Bytecode {
    pub bytes: Box<[u8]>
}

impl Bytecode {

    pub fn next_instruction(&self, offset: usize) -> (Instruction, usize) {
        let opcode = self.bytes[offset];

        match opcode {
            0x00 => (Instruction::NoOperation(), 1),
            0x01 => (Instruction::StoreIntoIntArray(), 1),
            0x02 => (Instruction::PushConstIntM1(), 1),
            0x03 => (Instruction::PushConstInt0(), 1),
            0x04 => (Instruction::PushConstInt1(), 1),
            0x05 => (Instruction::PushConstInt2(), 1),
            0x06 => (Instruction::PushConstInt3(), 1),
            0x07 => (Instruction::PushConstInt4(), 1),
            0x08 => (Instruction::PushConstInt5(), 1),
            0x0E => (Instruction::PushConstDouble0(), 1),
            0x0F => (Instruction::PushConstDouble1(), 1),

            0x10 => (Instruction::LoadByteImmediate(self.bytes[offset+1]), 2),
            0x11 => (Instruction::LoadShortImmediate((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0x12 => (Instruction::LoadConstant(self.bytes[offset+1]), 2),
            0x14 => (Instruction::LoadConstant64((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),

            0x1A => (Instruction::LoadLocalInt0(), 1),
            0x1B => (Instruction::LoadLocalInt1(), 1),
            0x1C => (Instruction::LoadLocalInt2(), 1),
            0x1D => (Instruction::LoadLocalInt3(), 1),
            0x26 => (Instruction::LoadLocalDouble0(), 1),
            0x27 => (Instruction::LoadLocalDouble1(), 1),
            0x28 => (Instruction::LoadLocalDouble2(), 1),
            0x29 => (Instruction::LoadLocalDouble3(), 1),
            0x2A => (Instruction::LoadLocalReference0(), 1),
            0x2B => (Instruction::LoadLocalReference1(), 1),
            0x2C => (Instruction::LoadLocalReference2(), 1),
            0x2D => (Instruction::LoadLocalReference3(), 1),

            0x3B => (Instruction::StoreLocalInt0(), 1),
            0x3C => (Instruction::StoreLocalInt1(), 1),
            0x3D => (Instruction::StoreLocalInt2(), 1),
            0x3E => (Instruction::StoreLocalInt3(), 1),
            0x4B => (Instruction::StoreReference0(), 1),
            0x4C => (Instruction::StoreReference1(), 1),
            0x4D => (Instruction::StoreReference2(), 1),
            0x4E => (Instruction::StoreReference3(), 1),

            0x57 => (Instruction::Pop(), 1),
            0x59 => (Instruction::Duplicate(), 1),

            0x68 => (Instruction::MultiplyInt(), 1),
            0x6D => (Instruction::DivideLong(), 1),

            0x7A => (Instruction::ShiftIntRight(), 1),

            0x80 => (Instruction::OrInt(), 1),

            0xAC => (Instruction::ReturnInt(), 1),

            0xB0 => (Instruction::ReturnReference(), 1),
            0xB1 => (Instruction::ReturnVoid(), 1),
            0xB2 => (Instruction::GetStatic((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB3 => (Instruction::PutStatic((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB4 => (Instruction::GetField((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB5 => (Instruction::PutField((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB6 => (Instruction::InvokeVirtual((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB7 => (Instruction::InvokeSpecial((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xB8 => (Instruction::InvokeStatic((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            0xBA => (Instruction::InvokeDynamic((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16, (self.bytes[offset+3] as u16) << 8 | self.bytes[offset+4] as u16), 5),
            0xBB => (Instruction::NewObject((self.bytes[offset+1] as u16) << 8 | self.bytes[offset+2] as u16), 3),
            _ => (Instruction::Unknown(opcode), 1)
        }
    }

    pub fn instructions(&self) -> Box<[Instruction]> {
        let mut v = Vec::with_capacity(self.bytes.len());
        
        let mut i = 0;
        while i < self.bytes.len() {

            let (instruction, offset) = self.next_instruction(i);

            v.push(instruction);
            i += offset;
        }

        v.into_boxed_slice()
    }
}

impl Debug for Bytecode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        f.debug_list()
            .entries(self.instructions())
            .finish()
    }
}

#[derive(Debug)]
#[repr(u8)]
pub enum Instruction {
    NoOperation()           = 0x00, // No-Operation
    StoreIntoIntArray()     = 0x01, // ..., arrayref, index, value
    PushConstIntM1()        = 0x02, // Push -1
    PushConstInt0()         = 0x03, // Push 0
    PushConstInt1()         = 0x04, // Push 1
    PushConstInt2()         = 0x05, // Push 2
    PushConstInt3()         = 0x06, // Push 3
    PushConstInt4()         = 0x07, // Push 4
    PushConstInt5()         = 0x08, // Push 5
    PushConstDouble0()      = 0x0E, // Push 0.0
    PushConstDouble1()      = 0x0F, // Push 1.0

    LoadByteImmediate(u8)   = 0x10, // push immediate short
    LoadShortImmediate(u16) = 0x11, // push immediate short
    LoadConstant(u8)        = 0x12, // Push from constant pool
    LoadConstant64(u16)     = 0x14, // Push Long or Double from constant pool
    LoadLocalInt0()         = 0x1A, // Load int from local variable
    LoadLocalInt1()         = 0x1B, // Load int from local variable
    LoadLocalInt2()         = 0x1C, // Load int from local variable
    LoadLocalInt3()         = 0x1D, // Load int from local variable

    LoadLocalDouble0()      = 0x26, // Load local double variable reference onto stack
    LoadLocalDouble1()      = 0x27, // Load local double variable reference onto stack
    LoadLocalDouble2()      = 0x28, // Load local double variable reference onto stack
    LoadLocalDouble3()      = 0x29, // Load local double variable reference onto stack
    LoadLocalReference0()   = 0x2A, // Load local reference variable reference onto stack
    LoadLocalReference1()   = 0x2B, // Load local reference variable reference onto stack
    LoadLocalReference2()   = 0x2C, // Load local reference variable reference onto stack
    LoadLocalReference3()   = 0x2D, // Load local reference variable reference onto stack

    StoreLocalInt0()        = 0x3B, // store int into local variable
    StoreLocalInt1()        = 0x3C, // store int into local variable
    StoreLocalInt2()        = 0x3D, // store int into local variable
    StoreLocalInt3()        = 0x3E, // store int into local variable
    StoreReference0()       = 0x4B, // store reference into local variable
    StoreReference1()       = 0x4C, // store reference into local variable
    StoreReference2()       = 0x4D, // store reference into local variable
    StoreReference3()       = 0x4E, // store reference into local variable

    Pop()                   = 0x57, // Pop top stack value
    Duplicate()             = 0x59, // duplicate top stack value

    MultiplyInt()           = 0x68, // int multiplication
    DivideLong()            = 0x6D, // long division

    ShiftIntRight()         = 0x7a, // shift int

    OrInt()                 = 0x80, // value, value => or

    ReturnInt()             = 0xAC, // return integer from function

    ReturnReference()       = 0xB0, // return top-ref from current function
    ReturnVoid()            = 0xB1, // return void from function
    GetStatic(u16)          = 0xB2, // get static field from class
    PutStatic(u16)          = 0xB3, // set static field on class
    GetField(u16)           = 0xB4, // get field from class
    PutField(u16)           = 0xB5, // set field to a value
    InvokeVirtual(u16)      = 0xB6, // invoke function on a class
    InvokeSpecial(u16)      = 0xB7, // invoke instance method
    InvokeStatic(u16)       = 0xB8, // invoke static function
    InvokeDynamic(u16, u16) = 0xBA, // invoke dynamic function
    NewObject(u16)          = 0xBB, // Create a new object from a constant-pool reference
    Unknown(u8),
}
